import rclpy
from rclpy.node import Node

from rclpy.callback_groups import ReentrantCallbackGroup

from interaction_msgs.srv import Query
from interaction_msgs.srv import Task
from interaction_msgs.msg import UI2Opt
from interaction_msgs.msg import TaskOrder
from interaction_msgs.msg import Opt2UI


from .optimizers.bayesian_optimization import BayesianOptimization

from transitions.extensions.asyncio import AsyncMachine
import asyncio

import threading
import time
import os

import numpy as np


class BayesianOptimizationNode(Node):
    def __init__(self, event_loop):
        super().__init__('bayesian_optimization_node')
        self.event_loop = event_loop

        # region Parameters
        """
        self.acquisition_function_name = (self.declare_parameter('acquisition_function_name', 'EI')
                                          .get_parameter_value().string_value)
        self.nr_bo_steps = self.declare_parameter('bo_steps', 100).get_parameter_value().integer_value
        self.nr_dimensions = self.declare_parameter('nr_dimensions', 1).get_parameter_value().integer_value
        self.nr_policy_parameters = (self.declare_parameter('nr_policy_parameters', 100)
                                     .get_parameter_value().integer_value)
        """
        self.kernel_type = self.declare_parameter('kernel_type', 'Matern').get_parameter_value().string_value
        self.nr_init = self.declare_parameter('nr_init', 3).get_parameter_value().integer_value
        self.task_request_timeout = self.declare_parameter('task_request_timeout', 10).get_parameter_value().integer_value
        self.task_request_max_tries = self.declare_parameter('task_request_max_tries', 3).get_parameter_value().integer_value
        # endregion

        # Subscribers
        self.ui_sub = self.create_subscription(UI2Opt, 'interaction/ui_response', self.ui_callback, 10)
        self.order_sub = self.create_subscription(TaskOrder, 'interaction/order', self.task_order_callback, 10)

        # Publishers
        self.ui_pub = self.create_publisher(Opt2UI, 'interaction/ui_request', 10)

        # Service Clients
        self.query_client = self.create_client(Query, 'interaction/user_query')
        self.task_client = self.create_client(Task, 'interaction/task_srv')

        # Bayesian Optimization
        self.bayesian_optimization = None

        self.current_episodes = 0
        self.nr_bo_steps = 0
        self.nr_dimensions = 0
        self.nr_policy_parameters = 0
        self.acquisition_function_name = ''
        self.fixed_dimensions = None

        self.seed = None
        self.lower_bound = None
        self.upper_bound = None

        # State Machine
        states = [
            {'name': 'idle'},
            {'name': 'initialize_bo', 'on_enter': 'initialize_bo_fun'},
            {'name': 'user_query', 'on_enter': 'user_query_fun'},
            {'name': 'sample_x_next', 'on_enter': 'sample_x_next_fun'},
            {'name': 'waiting_for_user_response', 'on_enter': 'waiting_for_user_response_fun'},
            {'name': 'processing', 'on_enter': 'processing_fun'},
            {'name': 'waiting_for_task_response', 'on_enter': 'waiting_for_task_response_fun'}
        ]
        transitions = [
            {'trigger': 'order_received', 'source': 'idle', 'dest': 'initialize_bo'},
            {'trigger': 'initialization_finished', 'source': 'initialize_bo', 'dest': 'user_query'},
            {'trigger': 'non_interactive', 'source': 'user_query', 'dest': 'sampling_x_next'},
            {'trigger': 'sampling_finished', 'source': 'sampling_x_next', 'dest': 'processing'},
            {'trigger': 'interactive', 'source': 'user_query', 'dest': 'waiting_for_user_response'},
            {'trigger': 'user_response_received', 'source': 'waiting_for_user_response', 'dest': 'processing'},
            {'trigger': 'processing_finished', 'source': 'processing', 'dest': 'waiting_for_task_response'},
            {'trigger': 'task_response_received', 'source': 'waiting_for_task_response', 'dest': 'user_query'},
            {'trigger': 'order_completed', 'source': 'waiting_for_task_response', 'dest': 'idle'},
            {'trigger': 'abort', 'source': '*', 'dest': 'idle'}
        ]

        """
        Algo:
        A) idle -> order_received() -> initializing_bo:
            order_sub - receives order
            reset_bo with the order parameters
            trigger: order_received()

        for i=0:nr_episodes
        B) initializing_bo -> initialization_finished() -> user_query:
            for j = 0:nr_init
                x_next <- initial sampling
                send x_next to task node
                reward received from task node
            fit model
            trigger: initialization_finished()

        C1.1) user_query -> non_interactive() -> sampling_x_next
            service call for user_query
            if user_query is False:
                trigger: non_interactive()
        C1.2) sampling_x_next -> sampling_finished() -> processing
            sampling x_next with acqusition function
            trigger: sampling_finished()

        C2.1) user_query -> interactive() -> waiting_for_user
            service call for user_query
            if user_query is True:
                trigger: interactive()
        C2.2) waiting_for_user_response -> user_response_received() -> processing
            send x_best to task node
            send x_best to UI
            user adapts x_best
            (UI sends x_next to task node)
            receives x_next from UI
            trigger: user_response_received() 

        D) processing -> processing_finished() -> waiting_for_task_response
            sends x_next to task node
            trigger: processing_finished()

        E) waiting_for_task_response -> task_response_received() -> user_query
            if episode < nr_episodes:
                get results from reward_sub
                model fit with additional observation
                trigger: task_response_received()

        F) waiting_for_task_response -> order_completed() -> idle
            if episode = nr_episodes:
                completion_pub with the completed results
                trigger: order_completed()


        Additional Transitions:
        Abort: * -> abort() -> idle
        """

        self.machine = AsyncMachine(model=self, states=states,
                                    transitions=transitions, initial='idle',
                                    ignore_invalid_triggers=True)

    def reset_bo(self, fixed_dimensions=None, **kwargs):
        self.bayesian_optimization = BayesianOptimization(self.nr_bo_steps,
                                                          self.nr_dimensions,
                                                          self.nr_policy_parameters,
                                                          seed=self.seed,
                                                          fixed_dimensions=fixed_dimensions,
                                                          lower_bound=self.lower_bound,
                                                          upper_bound=self.upper_bound,
                                                          acquisition_function_name=self.acquisition_function_name,
                                                          kernel_type=self.kernel_type, **kwargs)

    # region Callback functions
    def task_order_callback(self, msg):
        self.current_episodes = 0
        self.nr_bo_steps = msg.nr_bo_steps
        self.nr_dimensions = msg.nr_dimensions
        self.nr_policy_parameters = msg.nr_policy_parameters
        self.seed = msg.seed
        self.fixed_dimensions = msg.fixed_dimensions
        self.acquisition_function_name = msg.acquisition_function_name

        asyncio.run_coroutine_threadsafe(self.order_received(), self.event_loop)

    def reward_callback(self, msg):
        pass

    def ui_callback(self, msg):
        pass

    async def send_task_request_with_retry(self, request):
        for attempt in range(self.task_request_max_tries):
            self.get_logger().info(f"Attempt {attempt + 1}, send task request with retry!")

            future = self.task_client.call_async(request)
            try:
                response = await asyncio.wait_for(future, self.task_request_timeout)
                return response
            except asyncio.TimeoutError:
                self.get_logger().warning(f"Task request timed out, resending...")

        self.get_logger().error(f"Max retries reached for task request: {self.task_request_max_tries}")
        return None
    # endregion

    # State Methods
    async def initializing_bo_fun(self):
        self.get_logger().info('Initializing bo...')

        self.reset_bo(fixed_dimensions=self.fixed_dimensions)

        request = Task.Request()
        request.reward_return = False
        request.nr_dim = self.nr_dimensions
        request.nr_parameter = self.nr_policy_parameters
        for i in range(self.nr_init):
            request.x_next = self.bayesian_optimization.sample_random_next_observation()

            response = await self.send_task_request_with_retry(request)
            if i < self.nr_init:
                self.bayesian_optimization.add_observation(response.reward, response.x_observed, fit=False)
            else:
                self.bayesian_optimization.add_observation(response.reward, response.x_observed)

        await self.initialization_finished()

    async def user_query_fun(self):
        self.get_logger().info('Deciding wheter to ask the user...')
        srv_msg = Query()
        future = self.query_client.call_async(srv_msg)
        try:
            response = await asyncio.wait_for(future, timeout=2)
            if response.interaction:
                await self.interactive()

            else:
                await self.non_interactive()

        except asyncio.TimeoutError:
            self.get_logger().info("Timeout for Query Service...")

    async def sample_x_next_fun(self):
        pass

    async def waiting_for_user_response_fun(self):
        self.get_logger().info('Waiting for user response...')

    async def processing_fun(self):
        pass

    async def waiting_for_task_response_fun(self):
        pass


def run_asyncio_loop():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    loop.run_forever()
    return loop  # Ensure you return the loop reference


def main():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    threading.Thread(target=loop.run_forever, daemon=True).start()

    # Initialize ROS and pass the event loop to the node
    rclpy.init()
    node = BayesianOptimizationNode(loop)

    # Now, the ROS 2 node has the correct event loop reference for asyncio operations
    rclpy.spin(node)

    node.destroy_node()
    rclpy.shutdown()
    loop.call_soon_threadsafe(loop.stop)


if __name__ == '__main__':
    main()
